from flask import Response

from pokie.constants import DI_APP
from pokie.http import HttpErrorHandler


class TestErrorHandler:
    def test_errorhandler(self, pokie_app):
        app = pokie_app.di.get(DI_APP)
        obj = HttpErrorHandler(pokie_app.di)

        err = obj.error_400(pokie_app, None)
        assert isinstance(err, Response)
        assert err.status == "400 BAD REQUEST"
        assert err.json["success"] == False

        err = obj.error_404(pokie_app, None)
        assert isinstance(err, Response)
        assert err.status == "404 NOT FOUND"
        assert err.json["success"] == False

        err = obj.error_405(pokie_app, None)
        assert isinstance(err, Response)
        assert err.status == "405 METHOD NOT ALLOWED"
        assert err.json["success"] == False

        err = obj.error_500(pokie_app, None)
        assert isinstance(err, Response)
        assert err.status == "500 INTERNAL SERVER ERROR"
        assert err.json["success"] == False



from typing import List, Optional

from rick_db import Repository

from pokie.contrib.auth.dto.token import UserTokenRecord


class UserTokenRepository(Repository):
    def __init__(self, db):
        super().__init__(db, UserTokenRecord)

    def find_by_token(self, token: str) -> Optional[UserTokenRecord]:
        result = self.fetch_where(
            [
                (UserTokenRecord.token, "=", token),
            ]
        )
        if not len(result):
            return None
        return result.pop(0)

    def find_by_user(self, id_user: int) -> List[UserTokenRecord]:
        return self.fetch_where(
            [
                (UserTokenRecord.user, "=", id_user),
            ]
        )

    def prune(self, now) -> List[UserTokenRecord]:
        return self.delete_where(
            [
                (UserTokenRecord.expires, "<", now),
            ]
        )



from typing import List, Optional

from rick.base import Di
from rick.mixin.injectable import Injectable
from rick.resource import CacheInterface

from pokie.cache import DummyCache
from pokie.contrib.auth.constants import CFG_AUTH_USE_CACHE
from pokie.contrib.auth.dto import AclRoleRecord, AclResourceRecord
from pokie.contrib.auth.repository.acl import AclRoleRepository, AclResourceRepository
from pokie.constants import DI_DB, DI_CACHE, DI_CONFIG, TTL_1D


class AclService(Injectable):
    KEY_ROLE = "acl:role:{}"
    KEY_ROLE_RESOURCE = "acl:role:{}:resources"
    KEY_USER_ROLES = "user:{}:roles"
    TTL = TTL_1D

    def __init__(self, di: Di):
        super().__init__(di)
        self.cache = DummyCache(di)
        if di.get(DI_CONFIG).get(CFG_AUTH_USE_CACHE, False):
            if di.has(DI_CACHE):
                self.cache = di.get(DI_CACHE)

    def get_user_roles(self, id_user: int) -> dict:
        """
        Retrieve roles associated with the user
        :param id_user:
        :return: dict[int, AclRoleRecord]
        """
        key = self.KEY_USER_ROLES.format(id_user)
        id_roles = self.cache.get(key)
        if id_roles is not None:
            result = {}
            for id_role in id_roles:
                result[id_role] = self.get_role(id_role)
            return result

        result = self.role_repository.map_result_id(
            self.role_repository.find_user_roles(id_user)
        )
        user_roles = list(result.keys())

        if len(user_roles) > 0:
            # only cache if it has data
            self.cache.set(key, user_roles, self.TTL)

        return result

    def get_user_resources(self, id_user: int) -> dict:
        """
        Retrieve resources associated with the user
        :param id_user:
        :return: dict[int, AclResourceRecord]
        """
        resources = {}
        key = self.KEY_USER_ROLES.format(id_user)
        id_roles = self.cache.get(key)
        if id_roles is not None:
            for id_role in id_roles:
                for res in self.list_role_resources(id_role):
                    resources[res.id] = res
            return resources

        user_roles = []
        for _, role in self.get_user_roles(id_user).items():
            user_roles.append(role.id)
            for res in self.list_role_resources(role.id):
                resources[res.id] = res

        if len(user_roles) > 0:
            # only cache if it has data
            self.cache.set(key, user_roles, self.TTL)
        return resources

    def list_role_resources(self, id_role: int) -> List[AclResourceRecord]:
        """
        Retrieve all resources for the given role
        :param id_role:
        :return:
        """
        key = self.KEY_ROLE_RESOURCE.format(id_role)
        resource_list = self.cache.get(key)
        if resource_list is not None:
            return resource_list

        resource_list = self.resource_repository.find_by_role(id_role)
        if len(resource_list) > 0:
            # only cache if it has data
            self.cache.set(key, resource_list, self.TTL)
        return resource_list

    def list_roles(self) -> List[AclRoleRecord]:
        """
        Retrieve all roles
        :return:
        """
        return self.role_repository.fetch_all_ordered(AclRoleRecord.id)

    def list_resources(self) -> List[AclResourceRecord]:
        """
        Retrieve all resources
        :return:
        """
        return self.resource_repository.fetch_all_ordered(AclResourceRecord.id)

    def get_role(self, id_role: int) -> Optional[AclRoleRecord]:
        """
        Get Acl Role Record
        :param id_role:
        :return:
        """
        key = self.KEY_ROLE.format(id_role)
        record = self.cache.get(key)
        if record is not None:
            return record

        record = self.role_repository.fetch_pk(id_role)
        if record:
            self.cache.set(key, record, self.TTL)
        return record

    def get_resource(self, id_resource: str) -> Optional[AclResourceRecord]:
        """
        Get Acl Resource Record
        :param id_resource:
        :return:
        """
        return self.resource_repository.fetch_pk(id_resource)

    def add_role(self, description: str) -> int:
        """
        Add a new Role
        :except IntegrityError
        :param description:
        :return:
        """
        record = AclRoleRecord(description=description)
        record.id = self.role_repository.insert_pk(record)
        if record.id:
            key = self.KEY_ROLE.format(record.id)
            self.cache.set(key, record, self.TTL)
        return record.id

    def add_role_resource(self, id_role: int, id_resource: int):
        """
        Add Acl Resource to Role
        :param id_role:
        :param id_resource:
        :return:
        """
        self.role_repository.add_role_resource(id_role, id_resource)
        self.cache.remove(self.KEY_ROLE.format(id_role))
        self.cache.remove(self.KEY_ROLE_RESOURCE.format(id_role))

    def list_role_user_id(self, id_role: int) -> List[int]:
        """
        Retrieve a list of UserRecord id's associated with the role
        :param id_role:
        :return:
        """
        return self.role_repository.list_role_user_id(id_role)

    def add_user_role(self, id_user: int, id_role: int):
        """
        Add Acl Role to user
        :param id_user:
        :param id_role:
        :return:
        """
        self.role_repository.add_user_role(id_user, id_role)
        self.cache.remove(self.KEY_USER_ROLES.format(id_user))

    def remove_user_role(self, id_user: int, id_role: int):
        """
        Remove Acl role from user
        :param id_user:
        :param id_role:
        :return:
        """
        self.role_repository.remove_user_role(id_user, id_role)
        self.cache.remove(self.KEY_USER_ROLES.format(id_user))

    def remove_role(self, id_role: int):
        """
        Removes a role
        May raise DB Exception due to foreign keys
        :param id_role:
        :return:
        """
        self.role_repository.delete_pk(id_role)
        self.cache.remove(self.KEY_ROLE.format(id_role))
        self.cache.remove(self.KEY_ROLE_RESOURCE.format(id_role))

    def can_remove_role(self, id_role: int) -> bool:
        """
        Returns true if Acl Role is not associated with any user
        :param id_role:
        :return:
        """
        return self.role_repository.can_remove(id_role)

    def truncate_role_resources(self, id_role: int):
        """
        Removes Acl Resources from role
        :param id_role:
        :return:
        """
        self.role_repository.truncate_resources(id_role)
        self.cache.remove(self.KEY_ROLE_RESOURCE.format(id_role))

    def truncate_role_users(self, id_role: int):
        """
        Removes all users from role
        :param id_role:
        :return:
        """
        user_list = self.list_role_user_id(id_role)
        self.role_repository.truncate_users(id_role)
        for id_user in user_list:
            self.cache.remove(self.KEY_USER_ROLES.format(id_user))

    def remove_role_resource(self, id_role: int, id_resource: int):
        """
        Removes a Acl Resource from a Acl Role
        :param id_role:
        :param id_resource:
        :return:
        """
        self.role_repository.remove_role_resource(id_role, id_resource)
        self.cache.remove(self.KEY_ROLE_RESOURCE.format(id_role))

    @property
    def role_repository(self):
        return AclRoleRepository(self.get_di().get(DI_DB))

    @property
    def resource_repository(self):
        return AclResourceRepository(self.get_di().get(DI_DB))



import uuid
from rick.resource.config import StrOrFile


class PokieConfig:
    # default HTTP Exception Handler - 404 and 500 exceptions
    HTTP_ERROR_HANDLER = "pokie.http.HttpErrorHandler"

    # if true, all endpoints are authenticated by default
    USE_AUTH = True

    # Authentication provider
    AUTH_PROVIDER = "JWTProvider"

    # If true, ACL are loaded for users
    AUTH_ACL = True

    # Secret key for flask-login hashing
    AUTH_SECRET = uuid.uuid4().hex

    # Enables cache on User and Acl Services
    AUTH_USE_CACHE = True

    # cache table-related metadata (such as primary key info)
    # development should be false
    DB_CACHE_METADATA = False

    # Postgresql Configuration
    DB_NAME = "pokie"
    DB_HOST = "localhost"
    DB_PORT = 5432
    DB_USER = StrOrFile("postgres")
    DB_PASSWORD = StrOrFile("")
    DB_SSL = True
    DB_MINPROCS = 5
    DB_MAXPROCS = 15

    # Redis Configuration
    REDIS_HOST = "localhost"
    REDIS_PORT = 6379
    REDIS_PASSWORD = StrOrFile("")
    REDIS_DB = 0
    REDIS_SSL = "1"

    # Pytest Configuration
    TEST_DB_NAME = "pokie_test"  # test database parameters
    TEST_DB_HOST = "localhost"
    TEST_DB_PORT = 5432
    TEST_DB_USER = StrOrFile("postgres")
    TEST_DB_PASSWORD = StrOrFile("")
    TEST_DB_SSL = False

    TEST_MANAGE_DB = (
        False  # if false, unit testing does not manage db creation/migration
    )
    TEST_SHARE_CTX = False  # if false, each test has a separate context
    TEST_DB_REUSE = False  # if true, database is not dropped/recreated
    TEST_SKIP_MIGRATIONS = False  # if true, migrations are not run when recreating db
    TEST_SKIP_FIXTURES = False  # if true, fixtures are not run when recreating db



from pokie.constants import DI_DB
from rick.base import Di
from rick.mixin import Injectable, Runnable

from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


class ExampleFixture(Injectable, Runnable):
    def run(self, di: Di):
        db = di.get(DI_DB)
        repo = CustomerRepository(db)

        # insert a customer with "FIXTURE" as id
        if repo.fetch_pk("FIXTURE") is None:
            record = CustomerRecord(
                id="FIXTURE",
                company_name="company_name",
                contact_name="contact_name",
                contact_title="contact_title",
                address="address",
                city="city",
                region="region",
                postal_code="pc",
                country="country",
                phone="phone",
                fax="fax",
            )
            repo.insert(record)



from pokie.http import PokieView, ResponseRendererInterface


class HamburgerResponse(ResponseRendererInterface):
    def assemble(self, _app, **kwargs):
        # our custom Response only returns "hamburger"
        return "hamburger"


class CustomResponseView(PokieView):
    # custom response class to be used, intead of the default one
    response_class = HamburgerResponse

    def get(self):
        # just generate a response
        return self.success()



import pickle

from rick.base import Di
from rick.mixin import Injectable
from rick.resource import CacheInterface

from pokie.contrib.auth.dto import UserRecord


class MemoryCache(CacheInterface, Injectable):
    """
    In-memory cache

    Note: This cache implementation is for unit testing purposes only; DO NOT use it for regular development or
     production!!!
    """

    def __init__(self, di: Di):
        super().__init__(di)
        self.cache = {}

    def get(self, key):
        if key not in self.cache.keys():
            return None
        return pickle.loads(self.cache.get(key))

    def set(self, key, value, ttl=None):
        self.cache[key] = pickle.dumps(value)

    def has(self, key):
        return key in self.cache.keys()

    def remove(self, key):
        if key in self.cache.keys():
            del self.cache[key]

    def purge(self):
        self.cache = {}



from rick.base import Di
from rick.mixin import Injectable
from rick_db import Repository

from pokie.constants import DI_DB
from pokie.rest import RestServiceMixin
from pokie_test.dto import UsStatesRecord


class StatesService(Injectable, RestServiceMixin):
    record_class = UsStatesRecord
    repository_class = None  # optional custom repository class

    def __init__(self, di: Di):
        super().__init__(di)
        self._record_cls = self.record_class
        self._repository_cls = self.repository_class

    @property
    def repository(self) -> Repository:
        if self._record_cls is None:
            raise RuntimeError("Missing record class for repository")
        if self._repository_cls is None:
            return Repository(self.get_di().get(DI_DB), self._record_cls)
        else:
            return self._repository_cls(self.get_di().get(DI_DB))



from abc import ABC, abstractmethod
from typing import Any


class UserInterface(ABC):

    @abstractmethod
    def is_active(self):
        pass

    @abstractmethod
    def is_anonymous(self) -> bool:
        pass

    @abstractmethod
    def is_authenticated(self) -> bool:
        pass

    @abstractmethod
    def get_id(self):
        pass

    @abstractmethod
    def get_record(self):
        pass

    @abstractmethod
    def can_access(self, id_resource) -> bool:
        pass

    @abstractmethod
    def has_role(self, id_role):
        pass

    @abstractmethod
    def get_roles(self):
        pass

    @abstractmethod
    def get_resources(self):
        pass


class User(UserInterface):

    def __init__(
        self,
        id_user: Any = None,
        record: object = None,
        roles: list = None,
        resources: list = None,
        **kwargs
    ):
        self.id = id_user
        self.record = record
        if roles is None:
            roles = []
        self.roles = roles
        if resources is None:
            resources = []
        self.resources = resources
        # set custom parameters
        for k, v in kwargs:
            setattr(self, k, v)

    @property
    def is_active(self):
        if self.record is None:
            return self.id is not None
        active = getattr(self.record, "active", False)
        return active

    @property
    def is_anonymous(self) -> bool:
        return self.id is None

    @property
    def is_authenticated(self) -> bool:
        return self.id is not None

    def can_access(self, id_resource) -> bool:
        return id_resource in self.resources

    def has_role(self, id_role):
        return id_role in self.roles

    def get_id(self):
        if self.id:
            return str(self.id)
        return None

    def get_record(self):
        return self.record

    def get_roles(self):
        return self.roles

    def get_resources(self):
        return self.roles



from pokie.constants import HTTP_OK, HTTP_BADREQ
from pokie.test import PokieClient


class TestView:
    def test_requestrecord(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # GET should not use request deserialization
            result = client.get("/views/custom-requestrecord")
            assert result.code == HTTP_OK
            assert result.success is True

            # PUT must use request deserialization
            # empty values, should fail
            result = client.put("/views/custom-requestrecord", data={})
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # invalid values, should fail
            result = client.put("/views/custom-requestrecord", data={"field1": "abcd"})
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # valid values, should pass
            result = client.put(
                "/views/custom-requestrecord", data={"field1": "123", "field2": "abc"}
            )
            assert result.code == HTTP_OK
            assert result.success is True

            # POST must use request deserialization
            # empty values, should fail
            result = client.post("/views/custom-requestrecord", data={})
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # invalid values, should fail
            result = client.post("/views/custom-requestrecord", data={"field1": "abcd"})
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # valid values, should pass
            result = client.post(
                "/views/custom-requestrecord", data={"field1": "123", "field2": "abc"}
            )
            assert result.code == HTTP_OK
            assert result.success is True

            # PATCH must use request deserialization
            # empty values, should fail
            result = client.patch("/views/custom-requestrecord", data={})
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # invalid values, should fail
            result = client.patch(
                "/views/custom-requestrecord", data={"field1": "abcd"}
            )
            assert result.code == HTTP_BADREQ
            assert result.success is False
            assert isinstance(result.error, dict)
            for name in ["field1", "field2"]:
                assert name in result.form_error.keys()

            # valid values, should pass
            result = client.patch(
                "/views/custom-requestrecord", data={"field1": "123", "field2": "abc"}
            )
            assert result.code == HTTP_OK
            assert result.success is True

    def test_response(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # because of custom response, we need a custom handler
            result = client.client.get(
                "{}{}".format(client.base_url, "/views/custom-response")
            )
            assert result.status_code == HTTP_OK
            assert result.data == b"hamburger"

    def test_dispatch_hook(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # fetch view
            result = client.get("/views/hooks")
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data == "the quick brown fox jumps over the lazy dog"

    def test_camel_case(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # fetch view
            result = client.get("/views/camelcase")
            assert result.code == HTTP_OK
            assert result.success is True
            for name in ["companyName", "contactName", "contactTitle"]:
                assert name in result.data.keys()



from pokie.constants import HTTP_OK, HTTP_BADREQ
from pokie.rest import RestService
from pokie.test import PokieClient
from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


class TestRestService:
    def test_service(self, pokie_app):
        # build RestService based on pokie_test Customer
        svc = RestService(pokie_app.di)
        svc.set_record_class(CustomerRecord)
        svc.set_repository_class(CustomerRepository)

        assert isinstance(svc.repository, CustomerRepository)
        assert svc._record_cls == CustomerRecord

        # existing record
        assert svc.get("FIXTURE") is not None
        # non-existing record
        assert svc.get("ABCD") is None

        my_key = "ABCD"
        record = CustomerRecord(
            id=my_key,
            company_name="company_name",
            contact_name="contact_name",
            contact_title="contact_title",
            address="address",
            city="city",
            region="region",
            postal_code="pc",
            country="country",
            phone="phone",
            fax="fax",
        )
        # create
        assert svc.exists(my_key) is False
        assert svc.insert(record) == my_key
        assert svc.exists(my_key) is True

        # update key from ABCD to DEF
        record.company_name = "XYZ"
        svc.update(my_key, record)
        tmp = svc.get(my_key)
        assert tmp is not None
        assert tmp.company_name == record.company_name

        # delete
        svc.delete(my_key)
        assert svc.exists(my_key) is False

        # list
        total, rows = svc.list()
        assert total is not None
        assert total > 0
        assert len(rows) > 50



from rick.form import RequestRecord, field

from pokie.http import PokieView
from pokie_test.constants import SVC_NORTHWIND_CUSTOMER
from pokie_test.service import CustomerService


# RequestRecord class
class CustomerRequest(RequestRecord):
    fields = {
        "id": field(validators="required|maxlen:5"),
        "company_name": field(validators="required|maxlen:40"),
        "contact_name": field(validators="maxlen:30"),
        "contact_title": field(validators="maxlen:30"),
        "address": field(validators="maxlen:60"),
        "city": field(validators="maxlen:15"),
        "region": field(validators="maxlen:15"),
        "postal_code": field(validators="maxlen:15"),
        "country": field(validators="maxlen:15"),
        "phone": field(validators="maxlen:24"),
        "fax": field(validators="maxlen:24"),
    }


class CustomerController(PokieView):
    def view_customer(self, id_customer: str):
        """
        Controller-style GET endpoint

        :param id_customer:
        :return:
        """
        record = self.svc_customer().get_customer(id_customer)
        if not record:
            return self.not_found()

        return self.success(record)

    def svc_customer(self) -> CustomerService:
        return self.get_service(SVC_NORTHWIND_CUSTOMER)



import pytest
from psycopg2.errors import UniqueViolation, ForeignKeyViolation

from pokie.cache.memory import MemoryCache
from pokie.constants import DI_CACHE
from pokie.contrib.auth.constants import SVC_ACL, SVC_USER
from pokie.contrib.auth.dto import AclResourceRecord, UserRecord
from pokie.contrib.auth.service import AclService, UserService


class TestAclService:
    def test_acl_role(self, pokie_service_manager):
        svc_role = pokie_service_manager.get(SVC_ACL)  # type: AclService
        svc_user = pokie_service_manager.get(SVC_USER)  # type: UserService

        id_role = svc_role.add_role("role 1")
        assert id_role is not None
        role1 = svc_role.get_role(id_role)
        assert role1 is not None

        id_role = svc_role.add_role("role 2")
        assert id_role is not None
        role2 = svc_role.get_role(id_role)
        assert role2 is not None

        # test unique constraint
        with pytest.raises(UniqueViolation):
            svc_role.add_role("role 2")

        # add resource to role
        resource = AclResourceRecord(id="resource:foo", description="resource foo")
        svc_role.resource_repository.insert_pk(resource)

        # test unique constraint
        with pytest.raises(UniqueViolation):
            svc_role.resource_repository.insert_pk(resource)

        # add resource to role
        svc_role.add_role_resource(role1.id, resource.id)
        resources = svc_role.list_role_resources(role1.id)
        assert len(resources) == 1
        assert svc_role.can_remove_role(role1.id) is False

        resources = svc_role.list_role_resources(role2.id)
        assert len(resources) == 0
        assert svc_role.can_remove_role(role2.id) is True

        # remove resource from role
        svc_role.remove_role_resource(role1.id, resource.id)
        resources = svc_role.list_role_resources(role1.id)
        assert len(resources) == 0

        user = UserRecord(username="user1", password="")
        user.id = svc_user.add_user(user)
        assert user.id is not None

        # add user to role
        svc_role.add_user_role(user.id, role1.id)
        assert svc_role.can_remove_role(role1.id) is False

        users = svc_role.list_role_user_id(role1.id)
        assert len(users) == 1
        assert user.id in users

        # remove role
        svc_role.remove_role(role2.id)
        assert svc_role.get_role(role2.id) is None
        resources = svc_role.list_role_resources(role2.id)
        assert len(resources) == 0

        with pytest.raises(ForeignKeyViolation):
            svc_role.remove_role(role1.id)

        # remove user from role the hard way
        svc_role.truncate_role_users(role1.id)
        users = svc_role.list_role_user_id(role1.id)
        assert len(users) == 0
        # add it again
        svc_role.add_user_role(user.id, role1.id)

        # remove user from role
        svc_role.remove_user_role(user.id, role1.id)
        # remove role
        svc_role.remove_role(role1.id)
        assert svc_role.get_role(role1.id) is None

    def test_acl_resources(self, pokie_service_manager):
        svc_role = pokie_service_manager.get(SVC_ACL)  # type: AclService
        svc_user = pokie_service_manager.get(SVC_USER)  # type: UserService

        id_role = svc_role.add_role("role 1")
        assert id_role is not None
        role1 = svc_role.get_role(id_role)
        assert role1 is not None

        # create resources
        res1 = AclResourceRecord(id="resource:foo", description="resource foo")
        res2 = AclResourceRecord(id="resource:bar", description="resource bar")
        svc_role.resource_repository.insert_pk(res1)
        svc_role.resource_repository.insert_pk(res2)

        # add res1 to role
        svc_role.add_role_resource(role1.id, res1.id)

        # create user
        user = UserRecord(username="user1", password="")
        user.id = svc_user.add_user(user)
        assert user.id is not None

        # add user to role
        svc_role.add_user_role(user.id, role1.id)
        role_list = svc_role.get_user_roles(user.id)
        assert len(role_list) == 1
        assert user.id in role_list.keys()

        # some extra repository functions not exposed in service
        assert svc_role.resource_repository.can_remove(res1.id) is False
        assert svc_role.resource_repository.can_remove(res2.id) is True

        # list user resources
        resource_list = svc_role.get_user_resources(user.id)
        assert len(resource_list) == 1
        assert res1.id in resource_list.keys()

        # list user roles and resources again for cache hit
        role_list = svc_role.get_user_roles(user.id)
        assert len(role_list) == 1
        resource_list = svc_role.get_user_resources(user.id)
        assert len(resource_list) == 1
        assert res1.id in resource_list.keys()

        # truncate role resources
        svc_role.truncate_role_resources(role1.id)
        resource_list = svc_role.get_user_resources(user.id)
        assert len(resource_list) == 0


class TestCachedAclService(TestAclService):
    def test_cached_acl_role(self, pokie_di, pokie_service_manager):
        pokie_di.add(DI_CACHE, MemoryCache)
        self.test_acl_role(pokie_service_manager)

    def test_cached_acl_resources(self, pokie_di, pokie_service_manager):
        pokie_di.add(DI_CACHE, MemoryCache)
        self.test_acl_resources(pokie_service_manager)



from argparse import ArgumentParser
import getpass
from rick.form import RequestRecord, field

from pokie.contrib.auth.constants import SVC_USER
from pokie.constants import DI_SERVICES
from pokie.contrib.auth.dto import UserRecord
from pokie.contrib.auth.service import UserService
from pokie.core import CliCommand


class UserCreateRequest(RequestRecord):
    fields = {
        "username": field(
            validators="required|minlen:1|maxlen:200", error="invalid username"
        ),
        "email": field(
            validators="required|minlen:3|maxlen:200|email",
            error="invalid email address",
        ),
        "first_name": field(validators="maxlen:100", error="invalid first name"),
        "last_name": field(validators="maxlen:100", error="invalid last name"),
    }


class UserInfoRequest(RequestRecord):
    fields = {
        "username": field(
            validators="required|pk:user,username", error="invalid username"
        )
    }


class UserCommand(CliCommand):
    @property
    def svc_user(self) -> UserService:
        return self.get_di().get(DI_SERVICES).get(SVC_USER)


class UserCreateCmd(UserCommand):
    description = "create a new user"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", type=str, help="User name to create")
        parser.add_argument("email", type=str, help="User email")
        parser.add_argument("-f", "--first_name", help="First name", required=False)
        parser.add_argument("-l", "--last_name", help="First name", required=False)
        parser.add_argument(
            "--enabled",
            default=True,
            action="store_true",
            help="Enable user (default True)",
        )
        parser.add_argument(
            "--admin",
            default=False,
            action="store_true",
            help="Create admin (default False)",
        )

    def run(self, args) -> bool:
        req = UserCreateRequest()
        if not req.is_valid({"username": args.username, "email": args.email}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        if self.svc_user.get_by_username(args.username) is not None:
            self.tty.write(self.tty.colorizer.red("Error: username already exists"))
            return False

        record = UserRecord(
            username=args.username,
            email=args.email,
            admin=args.admin,
            active=args.enabled,
            password="",
            first_name=args.first_name,
            last_name=args.last_name,
        )
        id = self.svc_user.add_user(record)
        self.tty.write(
            "Created user #{} with username '{}' and email '{}'".format(
                str(id), args.username, args.email
            )
        )

        return True


class UserInfoCmd(UserCommand):
    description = "list user details"

    fields_labels = {
        "id": "User Id",
        "username": "Username",
        "active": "Active",
        "admin": "Administrator",
        "first_name": "First name",
        "last_name": "Last name",
        "email": "Email adress",
        "creation_date": "Creation date",
        "last_login": "Last login",
        "attributes": "Attributes",
    }

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", help="User name to search")

    def run(self, args) -> bool:
        req = UserInfoRequest()
        if not req.is_valid({"username": args.username}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        record = self.svc_user.get_by_username(args.username)
        if record is None:
            self.tty.write(self.tty.colorizer.red("Error: username not found"))
            return False

        data = record.asdict()
        for f, label in self.fields_labels.items():
            print("{:<16}: {}".format(label, str(data[f])))

        return True


class UserModCmd(UserCommand):
    description = "change user details, including password"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", help="User name of account to change")
        parser.add_argument(
            "-p", "--password", action="store_true", help="Change password"
        )
        parser.add_argument(
            "-a", "--noadmin", action="store_true", help="Remove admin privileges"
        )
        parser.add_argument(
            "-A", "--admin", action="store_true", help="Set admin privileges"
        )
        parser.add_argument(
            "-d", "--disabled", action="store_true", help="Disable user"
        )
        parser.add_argument("-e", "--enabled", action="store_true", help="Enable user")

    def run(self, args) -> bool:
        req = UserInfoRequest()
        if not req.is_valid({"username": args.username}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        if (
            not args.password
            and not args.noadmin
            and not args.admin
            and not args.enabled
            and not args.disabled
        ):
            self.tty.write(
                self.tty.colorizer.red("Error: missing modification options")
            )
            return False

        if args.noadmin and args.admin:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: mutually exclusive options for admin privileges"
                )
            )
            return False

        if args.disabled and args.enabled:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: mutually exclusive options for account status"
                )
            )
            return False

        record = self.svc_user.get_by_username(args.username)
        if record is None:
            self.tty.write(self.tty.colorizer.red("Error: username not found"))
            return False

        changes = False
        if args.password:
            not_valid = True
            pwd = ""
            while not_valid:
                pwd = getpass.getpass("New password:")
                if len(pwd) == 0:
                    self.tty.write("operation aborted")
                    return False
                confirmation = getpass.getpass("Confirm password:")

                pwd = pwd.strip()
                confirmation = confirmation.strip()
                if pwd != confirmation:
                    self.tty.write(
                        self.tty.colorizer.red(
                            "password and confirmation don't match, try again or press return to leave"
                        )
                    )
                else:
                    not_valid = False

            if len(pwd) > 0:
                changes = True
                if self.svc_user.update_password(record.username, pwd):
                    # reload record
                    record = self.svc_user.get_by_username(args.username)
                    self.tty.write(self.tty.colorizer.green("> updated user password"))
                else:
                    self.tty.write(
                        self.tty.colorizer.yellow(
                            "> password update failed; maybe the auth backend doesn't support local password changes?"
                        )
                    )
                    return False

        if args.noadmin:
            if record.admin:
                changes = True
                self.tty.write(
                    self.tty.colorizer.green("> removed admin privileges from user")
                )
            record.admin = False

        if args.admin:
            if not record.admin:
                changes = True
                self.tty.write(
                    self.tty.colorizer.green("> added admin privileges to user")
                )
            record.admin = True

        if args.enabled:
            if not record.active:
                changes = True
                self.tty.write(self.tty.colorizer.green("> enabled user account"))
            record.active = True

        if args.disabled:
            if record.active:
                changes = True
                self.tty.write(self.tty.colorizer.green("> disabled user account"))
            record.active = False

        if changes:
            self.svc_user.update_user(record)
            self.tty.write("User account updated successfully")
        else:
            self.tty.write("User account already conforms to required changes")
        return True


class UserListCmd(UserCommand):
    description = "list users"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("-o", "--offset", type=int, help="Start offset", default=0)
        parser.add_argument(
            "-c", "--count", type=int, help="Records to show", default=0
        )
        parser.add_argument(
            "-i", "--id", action="store_true", help="Sort by id", default=False
        )

    def run(self, args) -> bool:
        if args.offset < 0:
            self.tty.write(self.tty.colorizer.red("Error: offset cannot be negative"))
            return False

        if args.count < 0:
            self.tty.write(self.tty.colorizer.red("Error: count cannot be negative"))
            return False

        sort_field = None
        limit = 100
        if not args.id:
            sort_field = UserRecord.username
        if args.count != 0:
            limit = args.count
        details = self.svc_user.list_users(args.offset, limit, sort_field)
        # update limit var for display purposes
        if args.count == 0:
            limit = details[0]
        self.tty.write(
            self.tty.colorizer.green(
                "Displaying {} of {} users:".format(limit, details[0])
            )
        )

        for user in details[1]:  # type: UserRecord
            self.tty.write(
                "{:<12}: {}, {}, {} {}".format(
                    user.id, user.username, user.email, user.first_name, user.last_name
                )
            )

        return True



import pytest

from pokie.cache import MemoryCache, RedisCache


@pytest.fixture
def redis_cache(pokie_di):
    return RedisCache(pokie_di)


class TestRedisCache:

    def test_mutability(self, redis_cache):
        key = "key1"
        data = {"key": "value"}
        assert redis_cache.has(key) is False
        redis_cache.set(key, data)
        assert redis_cache.has(key) is True
        # mutate stored object
        data["key"] = 3

        # test for cache immutability
        record = redis_cache.get(key)
        assert record is not None
        assert record["key"] == "value"



import humps
from rick.form import RequestRecord, field

from pokie.codegen.pg import PgTableSpec
from pokie.codegen import RequestGenerator, RecordGenerator

record_tablespec = """
"""

record_tablespec_camel = """
"""

tablespec_serial_names = [
    "id_tablespec",
    "field_bigint",
    "field_bigserial",
    "field_bit",
    "field_varbit",
    "field_box",
    "field_bytea",
    "field_char",
    "field_varchar",
    "field_cidr",
    "field_circle",
    "field_date",
    "field_float8",
    "field_inet",
    "field_int4",
    "field_interval",
    "field_json",
    "field_jsonb",
    "field_line",
    "field_lseg",
    "field_macaddr",
    "field_macaddr8",
    "field_money",
    "field_numeric",
    "field_path",
    "field_point",
    "field_polygon",
    "field_float4",
    "field_int2",
    "field_smallserial",
    "field_text",
    "field_time",
    "field_timetz",
    "field_timestamp",
    "field_timestamptz",
    "field_tsquery",
    "field_tsvector",
    "field_uuid",
    "field_xml",
]


class TestRecordGenerator:
    def test_generate_source(self, pokie_db):
        pg_spec = PgTableSpec(pokie_db)
        spec = pg_spec.generate("tablespec_serial")
        generator = RecordGenerator()
        # default settings
        src = generator.generate_source(spec)
        assert self._cleanup(src) == self._cleanup(record_tablespec)

        # camelCase
        src = generator.generate_source(spec, camelcase=True)
        assert self._cleanup(src) == self._cleanup(record_tablespec_camel)

    def test_generate_class(self, pokie_db):
        pg_spec = PgTableSpec(pokie_db)
        spec = pg_spec.generate("tablespec_serial")
        generator = RecordGenerator()

        cls = generator.generate_class(spec)
        for name in tablespec_serial_names:
            if name == "id_tablespec":
                assert getattr(cls, "id") == name
            else:
                assert getattr(cls, name) == name

        cls = generator.generate_class(spec, camelcase=True)
        for name in tablespec_serial_names:
            if name == "id_tablespec":
                assert getattr(cls, "id") == name
            else:
                assert getattr(cls, humps.camelize(name)) == name

    def _cleanup(self, s: str):
        for c in ["\n", " "]:
            s = s.replace(c, "")
        return c



# HttpErrorHandler
#
# This class provides Flask with JSON-enabled global error handler responses for common errors, such as 400,404,405 and
# 500.
# The HTTP Error Handler class mechanism is pluggable, and a custom one can be provided instead of the default one.
# To override the default one, just change the HTTP_ERROR_HANDLER config setting to point to a new custom class that
# extends *Injectable*
#
from rick.base import Di
from rick.mixin import Injectable

from pokie.constants import (
    HTTP_BADREQ,
    HTTP_NOT_FOUND,
    HTTP_INTERNAL_ERROR,
    DI_FLASK,
    HTTP_NOT_ALLOWED,
)
from .response import JsonResponse


class HttpErrorHandler(Injectable):
    ERROR_400 = "400 Bad Request: The browser (or proxy) sent a request that this server could not understand."
    ERROR_404 = "404 Not Found: The requested URL was not found on the server."
    ERROR_405 = (
        "405 Method Not Allowed: The method is not allowed for the requested URL."
    )
    ERROR_500 = "500 Internal Server Error"

    def __init__(self, di: Di):
        super().__init__(di)
        _app = di.get(DI_FLASK)

        def wrapper_400(e):
            return self.error_400(_app, e)

        def wrapper_404(e):
            return self.error_404(_app, e)

        def wrapper_405(e):
            return self.error_500(_app, e)

        def wrapper_500(e):
            return self.error_500(_app, e)

        # register global error handler
        _app.register_error_handler(400, wrapper_400)
        _app.register_error_handler(404, wrapper_404)
        _app.register_error_handler(405, wrapper_405)
        _app.register_error_handler(500, wrapper_500)

    def error_400(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_400}, success=False, code=HTTP_BADREQ
        )
        return r.assemble(_app)

    def error_404(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_404}, success=False, code=HTTP_NOT_FOUND
        )
        return r.assemble(_app)

    def error_405(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_405}, success=False, code=HTTP_NOT_ALLOWED
        )
        return r.assemble(_app)

    def error_500(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_500}, success=False, code=HTTP_INTERNAL_ERROR
        )
        return r.assemble(_app)

    def response(self, **kwargs):
        return JsonResponse(**kwargs)



from pokie.cache import MemoryCache


class TestMemoryCache:

    def test_mutability(self, pokie_di):
        cache = MemoryCache(pokie_di)
        key = "key1"
        data = {"key": "value"}
        assert cache.has(key) is False
        cache.set(key, data)
        assert cache.has(key) is True
        # mutate stored object
        data["key"] = 3

        # test for cache immutability
        record = cache.get(key)
        assert record is not None
        assert record["key"] == "value"



from rick_db import fieldmapper


@fieldmapper(tablename="acl_role", pk="id_acl_role")
class AclRoleRecord:
    id = "id_acl_role"
    description = "description"


@fieldmapper(tablename="acl_resource", pk="id_acl_resource")
class AclResourceRecord:
    id = "id_acl_resource"
    description = "description"


@fieldmapper(tablename="acl_role_resource", pk="id_acl_role_resource")
class AclRoleResourceRecord:
    id = "id_acl_role_resource"
    id_role = "fk_acl_role"
    id_resource = "fk_acl_resource"


@fieldmapper(tablename="acl_user_role", pk="id_acl_user_role")
class AclUserRoleRecord:
    id = "id_acl_user_role"
    id_role = "fk_acl_role"
    id_user = "fk_user"



from flask_login import login_user, logout_user
from rick.event import EventManager

from pokie.constants import DI_EVENTS
from pokie.contrib.auth.constants import SVC_USER, SVC_AUTH, SVC_ACL
from pokie.contrib.auth.service import UserService, AuthService, AclService
from pokie.http.view import PokieView, PokieAuthView
from rick.form import field, RequestRecord


class LoginRequest(RequestRecord):
    fields = {
        "username": field(validators="required"),
        "password": field(validators="required"),
        "remember": field(validators="bool"),
    }


class LoginView(PokieView):
    request_class = LoginRequest

    def post(self):
        username = self.request.get("username")
        pwd = self.request.get("password")
        remember = bool(self.request.get("remember"))

        user = self.svc_auth().authenticate(username, pwd)
        if user is None:
            return self.error("invalid credentials")

        user.password = None

        result = self.svc_acl().get_user_acl_info(user.id)
        result["user"] = user.asdict()

        self.mgr_event().dispatch(self.di, "afterLogin", result)

        # flask-login
        login_user(user, bool(remember))

        return self.success(result)

    def mgr_event(self) -> EventManager:
        return self.di.get(DI_EVENTS)

    def svc_user(self) -> UserService:
        return self.get_service(SVC_USER)

    def svc_auth(self) -> AuthService:
        return self.get_service(SVC_AUTH)

    def svc_acl(self) -> AclService:
        return self.get_service(SVC_ACL)


class LogoutView(PokieAuthView):
    def get(self):
        logout_user()
        return self.success()



from rick_db import fieldmapper


@fieldmapper(tablename="customers", pk="customer_id")
class CustomerRecord:
    id = "customer_id"
    company_name = "company_name"
    contact_name = "contact_name"
    contact_title = "contact_title"
    address = "address"
    city = "city"
    region = "region"
    postal_code = "postal_code"
    country = "country"
    phone = "phone"
    fax = "fax"


@fieldmapper(tablename="categories", pk="category_id")
class CategoryRecord:
    id = "category_id"
    name = "category_name"
    description = "description"
    picture = "picture"


@fieldmapper(tablename="shippers", pk="shipper_id")
class ShipperRecord:
    id = "shipper_id"
    name = "company_name"
    phone = "phone"


@fieldmapper(tablename="us_states", pk="state_id")
class UsStatesRecord:
    id = "state_id"
    name = "state_name"
    abbr = "state_abbr"
    region = "state_region"


@fieldmapper(tablename="territories", pk="territory_id")
class TerritoriesRecord:
    id = "territory_id"
    territory_description = "territory_description"
    region_id = "region_id"



from rick.resource.config import EnvironmentConfig
from pokie.config import PokieConfig
from pokie.core import FlaskApplication
from pokie.core.factories.pgsql import PgSqlFactory
from pokie.contrib.base.module import Module


class Config(EnvironmentConfig, PokieConfig):
    # ENV var with optional plugins
    POKIE_PLUGINS = []

    TEST_MANAGE_DB = True
    TEST_DB_SSL = False


def main(args=None):
    # patch base module to only allow some commands
    allowed_cmds = [
        "list",
        "help",
        "version",
        "db:init",
        "db:check",
        "codegen:dto",
        "codegen:request",
        "codegen:module",
        "codegen:app",
    ]
    cli_list = {}
    for cmd, cls in getattr(Module, "cmd", {}).items():
        if cmd in allowed_cmds:
            cli_list[cmd] = cls
    setattr(Module, "cmd", cli_list)

    # build cli app
    cfg = Config().build()
    pokie_app = FlaskApplication(cfg)
    pokie_app.build(
        cfg.get("pokie_plugins"),
        [
            PgSqlFactory,
        ],
    )
    pokie_app.cli()



import secrets
from datetime import datetime, timezone, timedelta

import pytest
from psycopg2.errors import UniqueViolation

from pokie.cache.memory import MemoryCache
from pokie.constants import DI_CACHE
from pokie.contrib.auth.constants import SVC_USER
from pokie.contrib.auth.dto import UserRecord
from pokie.contrib.auth.service import UserService


class TestUserService:
    def test_user(self, pokie_service_manager):
        svc_user = pokie_service_manager.get(SVC_USER)  # type: UserService

        # create users
        user1 = UserRecord(username="user1", password="")
        user1.id = svc_user.add_user(user1)
        assert user1.id is not None

        # exception with duplicate username
        with pytest.raises(UniqueViolation):
            _ = svc_user.add_user(user1)

        user2 = UserRecord(username="user2", password="abc")
        user2.id = svc_user.add_user(user2)
        assert user1.id is not None

        # find by username
        record = svc_user.get_by_username(user2.username)
        assert record is not None
        assert record.id == user2.id

        # list users
        user_list = svc_user.list_users(0, 100)
        assert len(user_list) == 2

        pwd = svc_user.hasher.hash("somePassword")
        svc_user.update_password(user1.id, pwd)
        for record in [svc_user.get_by_id(user1.id), svc_user.get_by_username("user1")]:
            assert record is not None
            assert record.id == user1.id
            assert len(record.password) > 0

        # test authentication
        user1 = svc_user.get_by_id(user1.id)
        assert user1.last_login is None
        valid_user = svc_user.authenticate(user1.username, "somePassword")
        assert valid_user is not None
        assert valid_user.id == user1.id
        assert valid_user.password is None
        assert valid_user.last_login is not None

        invalid_user = svc_user.authenticate(user1.username, "someInvalidPassword")
        assert invalid_user is None

        # test load_id and sanitization
        record = svc_user.load_id(user1.id)
        assert record is not None
        assert record.password is None

        # test update user
        data = {"key": "value"}
        user1.first_name = "john"
        user1.last_name = "connor"
        user1.attributes = data
        svc_user.update_user(user1)
        record = svc_user.get_by_id(user1.id)
        assert len(record.attributes) == 1
        assert record.attributes["key"] == "value"
        assert record.first_name == "john"
        assert record.last_name == "connor"

    def test_user_token(self, pokie_service_manager):
        svc_user = pokie_service_manager.get(SVC_USER)  # type: UserService

        # create users
        user1 = UserRecord(username="user1", password="")
        user1.id = svc_user.add_user(user1)
        user2 = UserRecord(username="user2", password="")
        user2.id = svc_user.add_user(user2)

        # create tokens
        tok1 = svc_user.add_user_token(user1.id)
        assert tok1 is not None
        assert len(tok1.token) > 0
        assert tok1.expires is None
        assert tok1.user == user1.id
        assert tok1.active == True
        later = datetime.now(timezone.utc) + timedelta(minutes=10)
        tok2 = svc_user.add_user_token(user2.id, later)
        assert tok2 is not None
        assert len(tok2.token) > 0
        assert tok2.expires is not None
        assert tok2.user == user2.id
        assert tok2.active == True

        # find user by non-existing tokens
        for i in range(0, 10):
            user = svc_user.get_user_by_token(secrets.token_hex())
            assert user is None

        # find user from existing token
        user = svc_user.get_user_by_token(tok1.token)
        assert user is not None
        assert user.id == user1.id
        user = svc_user.get_user_by_token(tok2.token)
        assert user is not None
        assert user.id == user2.id

        # disable token
        assert svc_user.disable_user_token(4927) is False
        assert svc_user.disable_user_token(tok1.id) is True

        # token inactive, should fail
        user = svc_user.get_user_by_token(tok1.token)
        assert user is None
        tok = svc_user.get_token(tok1.token)
        assert tok.active is False

        # list tokens
        token_list = svc_user.list_user_tokens(user2.id)
        assert len(token_list) == 1
        assert token_list[0].user == user2.id

        # remove token
        user = svc_user.get_user_by_token(tok2.token)
        assert user is not None
        svc_user.remove_user_token(tok2.id)
        user = svc_user.get_user_by_token(tok2.token)
        assert user is None

        # add a new token, force expiry
        later = datetime.now(timezone.utc) + timedelta(minutes=10)
        tok1 = svc_user.add_user_token(user2.id, later)
        user = svc_user.get_user_by_token(tok1.token)
        assert user is not None

        # warning: this will cause cache inconsistency
        tok1.expires = datetime.now(timezone.utc) - timedelta(minutes=10)
        svc_user.user_token_repository.update(tok1)
        svc_user.prune_tokens()
        tok = svc_user.user_token_repository.find_by_token(tok1.token)
        assert tok is None


class TestCachedUserService(TestUserService):
    def test_cached_user(self, pokie_di, pokie_service_manager):
        pokie_di.add(DI_CACHE, MemoryCache(pokie_di))
        self.test_user(pokie_service_manager)

    def test_cached_user_token(self, pokie_di, pokie_service_manager):
        pokie_di.add(DI_CACHE, MemoryCache(pokie_di))
        self.test_user_token(pokie_service_manager)



from typing import List

from flask import request

from pokie.http import DbGridRequest, PokieView
from pokie.rest import RestService, RestServiceMixin
from pokie.constants import DI_SERVICES


class RestView(PokieView):
    record_class = None
    search_fields = None  # type: List
    service_name = None
    list_limit = -1
    camel_case = False

    def get(self, id_record=None):
        """
        Read single record by id
        :param id_record:
        :return:
        """
        if id_record is None:
            return self.list()

        record = self.svc.get(id_record)
        if record is None:
            return self.not_found()

        return self.success(record)

    def list(self):
        """
        Query records
        :return:
        """
        search_fields = self.search_fields if self.search_fields is not None else []
        dbgrid_request = DbGridRequest(
            self.record_class, use_camel_case=self.camel_case
        )

        if not dbgrid_request.is_valid(request.args):
            return self.request_error(dbgrid_request)
        try:
            count, data = self.svc.list(
                **dbgrid_request.dbgrid_parameters(self.list_limit, search_fields)
            )
            result = {"total": count, "items": data}
            return self.success(result)
        except Exception as e:
            # exception may happen because of mismatched data type, such as matching strings to int fields
            self.logger.exception(e)
            return self.error()

    def post(self):
        """
        Create Record
        :return:
        """
        record = self.request.bind(self.record_class)
        self.svc.insert(record)
        return self.success()

    def put(self, id_record):
        """
        Update Record
        :return:
        """
        record = self.request.bind(self.record_class)
        self.svc.update(id_record, record)
        return self.success()

    def delete(self, id_record):
        """
        delete Record by id
        :param id_record:
        :return:
        """
        if not self.svc.exists(id_record):
            return self.not_found()

        self.svc.delete(id_record)
        return self.success()

    @property
    def svc(self) -> RestService:
        mgr = self.di.get(DI_SERVICES)
        if not self.service_name:
            svc_name = "svc.rest.{}.{}".format(
                self.__module__,
                str(self.record_class.__name__).replace("Record", "", 1),
            )
            if mgr.contains(svc_name):
                return mgr.get(svc_name)

            # build service
            svc = RestService(self.di)
            svc.set_record_class(self.record_class)

            # register it in the service manager
            mgr.register(svc_name, svc)
            return svc

        svc = mgr.get(self.service_name)
        if not isinstance(svc, RestServiceMixin):
            raise RuntimeError("Service '{}' does not implement RestService mixin")
        return svc



from pokie.contrib.base.dto import FixtureRecord
from pokie.contrib.base.repository.fixture import FixtureRepository
from pokie_test.repository import CustomerRepository


class TestFixtures:
    def test_load_fixtures(self, pokie_db):
        # check if table/record exists
        repo = FixtureRepository(pokie_db)
        f_list = repo.fetch_where(
            [
                (FixtureRecord.name, "=", "pokie_test.fixtures.ExampleFixture"),
            ]
        )
        assert f_list is not None
        assert len(f_list) == 1

        # fixture should have been loaded
        repo = CustomerRepository(pokie_db)
        assert repo.fetch_pk("FIXTURE") is not None



from rick.base import Di
from rick.mixin import Injectable, Runnable


class SampleJob(Injectable, Runnable):
    KEY = "SAMPLE_JOB_KEY"

    def __init__(self, di: Di):
        super().__init__(di)
        self.counter = 0

    def run(self, di: Di):
        self.counter += 1
        di.add(self.KEY, self.counter, replace=di.has(self.KEY))



import secrets
from datetime import datetime, timezone
from typing import Optional, List

from rick.base import Di
from rick.crypto.hasher import HasherInterface
from rick.crypto.hasher.bcrypt import BcryptHasher
from rick.mixin import Injectable

from pokie.cache import DummyCache
from pokie.contrib.auth.constants import CFG_AUTH_USE_CACHE
from pokie.contrib.auth.repository import UserTokenRepository
from pokie.contrib.auth.repository.user import UserRepository
from pokie.constants import DI_DB, DI_CACHE, TTL_1D, DI_CONFIG
from rick.util.datetime import iso8601_now
from pokie.contrib.auth.dto import UserRecord, UserTokenRecord


class UserService(Injectable):
    KEY_USERNAME = "user:username:{}"
    KEY_USER = "user:{}"
    KEY_TOKEN = "user:token:{}"
    TTL = TTL_1D

    def __init__(self, di: Di):
        super().__init__(di)
        self.cache = DummyCache(di)
        if di.get(DI_CONFIG).get(CFG_AUTH_USE_CACHE, False):
            if di.has(DI_CACHE):
                self.cache = di.get(DI_CACHE)

    def authenticate(self, username: str, password: str) -> Optional[UserRecord]:
        """
        Attempts to authenticate a user

        :param username:
        :param password:
        :return:
        """
        record = self.get_by_username(username)
        if record is None:
            return None

        if not record.active:
            return None

        if self.hasher.is_valid(password, record.password):
            if self.hasher.need_rehash(record.password):
                # update weak password hash
                self.update_password(record.id, self.hasher.hash(password))

            # update lastlogin
            self.update_lastlogin(record.id)
            return self.load_id(record.id)
        return None

    def sanitize_record(self, record: UserRecord) -> UserRecord:
        """
        Remove sensitive attributes from a UserRecord
        :param record:
        :return:
        """
        record.password = None
        return record

    def load_id(self, id_user: int) -> Optional[UserRecord]:
        """
        Loads a UserRecord with the password cleared out
        :param id_user:
        :return:
        """
        record = self.get_by_id(id_user)
        if record:
            return self.sanitize_record(record)
        return None

    def get_by_id(self, id_user: int) -> Optional[UserRecord]:
        """
        Find user by id
        :param id_user:
        :return:
        """
        key = self.KEY_USER.format(id_user)
        record = self.cache.get(key)
        if record:
            return record

        record = self.user_repository.fetch_pk(id_user)
        if record:
            self.cache.set(key, record, self.TTL)
        return record

    def get_by_username(self, username: str) -> Optional[UserRecord]:
        """
        Find user by username
        :param username:
        :return:
        """
        key = self.KEY_USERNAME.format(username)
        id_user = self.cache.get(key)
        if id_user is not None:
            return self.get_by_id(id_user)

        record = self.user_repository.find_by_username(username)
        if not record:
            return None

        # store username -> id map
        self.cache.set(key, record.id, self.TTL)

        # store user
        key = self.KEY_USER.format(record.id)
        self.cache.set(key, record, self.TTL)
        return record

    def update_lastlogin(self, id_user: int):
        """
        Update user last login timestamp
        :param id_user:
        :return:
        """
        now = datetime.now(timezone.utc)
        self.user_repository.update(UserRecord(id=id_user, last_login=now))
        key = self.KEY_USER.format(id_user)
        record = self.cache.get(key)
        if record:
            record.last_login = now
            self.cache.set(key, record, self.TTL)

    def update_password(self, id_user: int, password_hash: str):
        """
        Update user password
        :param id_user:
        :param password_hash:
        :return:
        """
        self.user_repository.update(UserRecord(id=id_user, password=password_hash))
        self.cache.remove(self.KEY_USER.format(id_user))

    def add_user(self, record: UserRecord) -> int:
        """
        Creates a new user
        :param record:
        :return:
        """
        return self.user_repository.insert_pk(record)

    def list_users(self, offset, limit, sort_field=None, sort_order=None) -> tuple:
        """
        Returns a tuple with a list of users

        :param offset:
        :param limit:
        :param sort_field:
        :param sort_order:
        :return: (total user count, [records])
        """
        return self.user_repository.list_users(offset, limit, sort_field, sort_order)

    def update_user(self, record: UserRecord):
        """
        Update UserRecord
        :param record:
        :return:
        """
        self.user_repository.update(record)
        self.cache.remove(self.KEY_USER.format(record.id))

    def get_user_by_token(self, token: str) -> Optional[UserRecord]:
        """
        Find UserRecord by token
        Expired tokens and inactive tokens are ignored

        :param token:
        :return:
        """
        now = datetime.now(timezone.utc)
        key = self.KEY_TOKEN.format(token)
        record = self.cache.get(key)
        if not record:
            record = self.user_token_repository.find_by_token(token)
            if not record:
                return None
            self.cache.set(key, record, self.TTL)

        if not record.active:
            return None

        if record.expires:
            if record.expires < now:
                return None
        return self.get_by_id(record.user)

    def get_token(self, token: str) -> Optional[UserTokenRecord]:
        """
        Find UserTokenRecord by token
        :param token:
        :return:
        """
        return self.user_token_repository.find_by_token(token)

    def add_user_token(self, id_user: int, expires: datetime = None) -> UserTokenRecord:
        """
        Creates a new Token for the given user
        :param id_user:
        :param expires:
        :return:
        """
        record = UserTokenRecord(
            creation_date=iso8601_now(),
            active=True,
            user=id_user,
            token=secrets.token_hex(64),
            expires=expires,
        )
        record.id = self.user_token_repository.insert_pk(record)
        return record

    def disable_user_token(self, id_user_token: int) -> bool:
        """
        Disable a UserTokenRecord
        :param id_user_token:
        :return:
        """
        record = self.user_token_repository.fetch_pk(id_user_token)
        if not record:
            return False
        key = self.KEY_TOKEN.format(record.token)
        if record.expires:
            if record.expires < datetime.now(timezone.utc).now():
                self.cache.remove(key)

        if not record.active:
            return True

        record.active = False
        self.user_token_repository.update(record)
        self.cache.set(key, record, self.TTL)
        return True

    def remove_user_token(self, id_user_token: int):
        """
        Removes a UserTokenRecord
        :param id_user_token:
        :return:
        """
        record = self.user_token_repository.fetch_pk(id_user_token)
        if not record:
            return None

        key = self.KEY_TOKEN.format(record.token)
        if self.cache.get(key):
            self.cache.remove(key)
        return self.user_token_repository.delete_pk(id_user_token)

    def list_user_tokens(self, id_user: int) -> List[UserTokenRecord]:
        """
        List UserTokenRecord for a given user
        :param id_user:
        :return:
        """
        return self.user_token_repository.find_by_user(id_user)

    def prune_tokens(self):
        """
        Remove all expired tokens
        :return:
        """
        now = datetime.now(timezone.utc)
        self.user_token_repository.prune(now)

    @property
    def user_repository(self) -> UserRepository:
        return UserRepository(self._di.get(DI_DB))

    @property
    def user_token_repository(self) -> UserTokenRepository:
        return UserTokenRepository(self._di.get(DI_DB))

    @property
    def hasher(self) -> HasherInterface:
        return BcryptHasher()



from pokie.constants import DI_APP
from pokie.contrib.base.cli.base import BaseCommand


class JobListCmd(BaseCommand):
    description = "list registered job workers"

    def run(self, args) -> bool:
        for name, jobs in self.get_di().get(DI_APP).get_jobs().items():
            self.tty.write("Worker Jobs for module {}:".format(name))
            for job in jobs:
                self.tty.write(
                    self.tty.colorizer.white("   {}".format(job), attr="bold")
                )

        return True


class JobRunCmd(BaseCommand):
    description = "run  all job workers"

    def run(self, args) -> bool:
        app = self.get_di().get(DI_APP)
        # run in loop
        app.job_runner()
        return True



from rick_db import fieldmapper


@fieldmapper(tablename="user_token", pk="id_user_token")
class UserTokenRecord:
    id = "id_user_token"
    creation_date = "creation_date"
    update_date = "update_date"
    active = "active"
    user = "fk_user"
    token = "token"
    expires = "expires"



import subprocess

from pokie.constants import DI_APP


class TestCli:
    """
    Tests CLI utils
    Note: mostly just test if clis actually run
    """

    def run_cmd(self, cmd: list, env: dict, retcode: int = 0):
        cmd = ["python3", "main.py"] + cmd
        if env:
            result = subprocess.run(cmd, capture_output=True, env=env)
        else:
            result = subprocess.run(cmd, capture_output=True)
        assert result.returncode == retcode

    # @todo: these tests are disabled because they currently do not detect the venv where they are run; as such,
    # no package dependencies actually exist, so all commands  fail
    #
    #    def test_list(self):
    #        self.run_cmd(["list"], {}, 0)
    #
    #    def test_help(self):
    #        self.run_cmd(["help", "version"], {}, 0)

    #    def test_version(self):
    #        self.run_cmd(["version"], {}, 0)

    #    def test_module(self, pokie_config):
    #        self.run_cmd(["module:list"], self.get_config(pokie_config))

    #    def test_route(self, pokie_config):
    #        self.run_cmd(["route:list"], self.get_config(pokie_config), 0)

    #    def test_db(self, pokie_config):
    #        self.run_cmd(["db:init"], self.get_config(pokie_config), 0)
    #        self.run_cmd(["db:check"], self.get_config(pokie_config), 0)
    #        self.run_cmd(["db:update"], self.get_config(pokie_config), 0)

    #    def test_job(self, pokie_config, pokie_di):
    #        self.run_cmd(["job:list"], self.get_config(pokie_config), 0)
    #        app = pokie_di.get(DI_APP)
    #        app.job_runner(single_run=True)
    # detect changes performed by SampleJob
    #        assert pokie_di.has("SAMPLE_JOB_KEY")
    #        assert pokie_di.get("SAMPLE_JOB_KEY") > 0

    def get_config(self, cfg) -> dict:
        result = {
            "DB_NAME": cfg.get("test_db_name"),
            "DB_HOST": cfg.get("test_db_host"),
            "DB_PORT": str(cfg.get("test_db_port")),
            "DB_USER": cfg.get("test_db_user"),
            "DB_PASSWORD": cfg.get("test_db_password"),
            "DB_SSL": str(cfg.get("test_db_ssl")),
        }
        return result

    def cleanup(self, s: str):
        for c in ["\n", " "]:
            s = s.replace(c, "")
        return c



from pokie.contrib.base.constants import SVC_SETTINGS
from pokie.contrib.base.service import SettingsService


class TestSettings:
    def test_settings(self, pokie_service_manager):
        svc = pokie_service_manager.get(SVC_SETTINGS)  # type: SettingsService

        assert len(svc.list()) == 0
        for i in range(1, 10):
            svc.upsert("module1", "setting_{}".format(i), str(i))
        assert len(svc.list()) == 9

        record = svc.by_key("module1", "setting_5")
        assert record is not None
        assert record.value == "5"

        svc.upsert(record.module, record.key, "foo")
        record = svc.by_key("module1", "setting_5")
        assert record is not None
        assert record.value == "foo"



from typing import Optional

from flask import Request
from rick.mixin import Injectable
import uuid

from flask_login import LoginManager, logout_user, login_user
from rick.base import Di

from pokie.constants import DI_CONFIG, DI_SERVICES, CFG_AUTH_SECRET, DI_FLASK
from pokie.contrib.auth import User
from pokie.contrib.auth.constants import SVC_USER, SVC_ACL
from pokie.contrib.auth.dto import UserRecord
from pokie.contrib.auth.service import AclService
from pokie.contrib.auth.user import UserInterface


def build_user_acl(di, user_record: UserRecord):
    svc_acl = di.get(DI_SERVICES).get(SVC_ACL)  # type: AclService
    roles = list(svc_acl.get_user_roles(user_record.id).keys())
    resources = list(svc_acl.get_user_resources(user_record.id).keys())
    return User(
        id_user=user_record.id, record=user_record, roles=roles, resources=resources
    )


class SessionProvider(Injectable):
    def __init__(self, di: Di):
        super().__init__(di)

        cfg = di.get(DI_CONFIG)
        app = di.get(DI_FLASK)
        app.secret_key = cfg.get(CFG_AUTH_SECRET, uuid.uuid4().hex)
        login_manager = LoginManager()
        login_manager.init_app(app)

        @login_manager.user_loader
        def load_user(user_id):
            # restores user profile from user service
            user_id = int(user_id)
            svc_manager = di.get(DI_SERVICES)
            user_record = svc_manager.get(SVC_USER).load_id(user_id)
            if not user_record:
                return None  # this will clear session
            return build_user_acl(di, user_record)

    def login(self, username, password, **kwargs) -> Optional[UserInterface]:
        svc_user = self.get_di().get(DI_SERVICES).get(SVC_USER)
        user_record = svc_user.authenticate(username, password)
        if not user_record:
            return None

        remember = False
        if "remember" in kwargs.keys():
            remember = bool(kwargs["remember"])
        user = build_user_acl(self.get_di(), user_record)

        # create user session
        login_user(user, remember)
        return user

    def get_user(self, req: Request, **kwargs) -> User:
        raise RuntimeError("SessionProvider: get_user() is not available directly")

    def logout(self, **kwargs):
        # remove user session
        logout_user()



from rick.base import Di
from rick.mixin import Injectable
from rick.resource import CacheInterface, CacheNull


class DummyCache(CacheNull, Injectable):
    """
    Dummy Cache Wrapper
    """

    def __init__(self, di: Di):
        super().__init__(di)



from typing import Type

from rick_db import Record
from rick.form import RequestRecord, field
from rick.mixin import Translator
import humps

from pokie.constants import DEFAULT_LIST_SIZE


class DbGridRequest(RequestRecord):
    FIELD_OFFSET = "offset"
    FIELD_LIMIT = "limit"
    FIELD_SORT = "sort"
    FIELD_MATCH = "match"
    FIELD_SEARCH = "search"

    fields = {
        FIELD_OFFSET: field(validators="numeric", value=0),
        FIELD_LIMIT: field(validators="numeric", value=DEFAULT_LIST_SIZE),
        FIELD_SORT: field(),
        FIELD_MATCH: field(),
        FIELD_SEARCH: field(),
    }

    def __init__(
        self, record: Type[Record], translator: Translator = None, use_camel_case=False
    ):
        super().__init__(translator)
        self.record = record
        self.use_camel_case = use_camel_case

    def _normalize(self, name) -> str:
        return humps.decamelize(name) if self.use_camel_case else name

    def validator_match(self, data, t: Translator):
        match_fields = data.get(self.FIELD_MATCH, None)
        if match_fields is not None:
            # if string is empty, ignore it
            if len(match_fields.strip()) == 0:
                match_fields = None
            else:
                match_fields = match_fields.split("|")
                # convert field names to column names
                result = {}
                for f in match_fields:
                    f = f.split(":", 1)
                    if len(f) != 2:
                        self.add_error(
                            self.FIELD_MATCH, t.t("invalid field match expression")
                        )
                        return False
                    name = self._normalize(f[0])
                    if name not in self.record._fieldmap.keys():
                        self.add_error(
                            self.FIELD_MATCH, t.t("invalid field name: {}").format(f)
                        )
                        return False
                    result[self.record._fieldmap[name]] = f[1]

                # replace original dict with result
                match_fields = result

        self.fields[self.FIELD_MATCH].value = match_fields

        return True

    def validator_sort(self, data, t: Translator):
        sort = data.get(self.FIELD_SORT, None)

        if sort is not None:
            sort = sort.split(",")
            result = {}
            # convert field, field:desc -> db_field:asc, db_field:desc
            for expr in sort:
                expr = expr.split(":")
                expr[0] = self._normalize(expr[0])
                if expr[0] not in self.record._fieldmap.keys():
                    self.add_error(
                        self.FIELD_SORT,
                        t.t("invalid sort field name: {}").format(expr[0]),
                    )
                    return False

                name = self.record._fieldmap[expr[0]]

                if len(expr) > 1:
                    if expr[1].lower() not in ["asc", "desc"]:
                        self.add_error(
                            self.FIELD_SORT,
                            t.t("invalid sort order: {}").format(expr[1]),
                        )
                        return False

                    result[name] = expr[1]
                else:
                    result[name] = "asc"

        self.fields[self.FIELD_SORT].value = sort
        return True

    def validator_offset(self, data, t: Translator):
        offset = data.get(self.FIELD_OFFSET, None)
        if offset is not None:
            offset = int(offset)
            if offset < 0:
                self.add_error(self.FIELD_OFFSET, t.t("invalid offset value"))
                return False
        self.fields[self.FIELD_OFFSET].value = offset
        return True

    def validator_limit(self, data, t: Translator):
        limit = data.get(self.FIELD_LIMIT, None)
        if limit is not None:
            limit = int(limit)
            if limit < 1:
                self.add_error(self.FIELD_LIMIT, t.t("invalid limit value"))
                return False

        self.fields[self.FIELD_LIMIT].value = limit
        return True

    def dbgrid_parameters(
        self, list_limit: int = 0, search_fields: list = None
    ) -> dict:
        """
        Return a list of parameters to be used as argument for DbGrid.run()
        :param list_limit:
        :param search_fields:
        :return:
        """
        offset = self.fields[self.FIELD_OFFSET].value
        limit = self.fields[self.FIELD_LIMIT].value

        # automatically cap records
        if offset is None and limit is None and list_limit > 0:
            limit = list_limit

        return {
            "search_text": self.fields[self.FIELD_SEARCH].value,
            "match_fields": self.fields[self.FIELD_MATCH].value,
            "limit": limit,
            "offset": offset,
            "sort_fields": self.fields[self.FIELD_SORT].value,
            "search_fields": search_fields,
        }



from tabulate import tabulate

from pokie.constants import DI_FLASK, DI_APP
from pokie.core import CliCommand


class RouteListCmd(CliCommand):
    description = "list routes"

    def run(self, args) -> bool:
        pokie_app = self.get_di().get(DI_APP)
        pokie_app.init()
        app = self.get_di().get(DI_FLASK)

        with app.app_context():
            result = []
            for rule in app.url_map.iter_rules():
                methods = ",".join(rule.methods)
                result.append([str(rule), methods])

            self.tty.write(tabulate(result, headers=["Endpoint", "Methods"]))
        return True



from rick.base import Di
from rick.resource.redis import RedisCache

from pokie.constants import DI_CACHE


def CacheFactory(_di: Di):
    """
    Cache factory
    Builds a CacheInterface object
    Currently only supports REDIS
    :param _di:
    :return:
    """
    _di.add(DI_CACHE, RedisCache)



from pokie.codegen.pg import PgTableSpec


class TestPgSpec:
    tablespec_fieldnames = {
        "id_tablespec": {
            "pk": True,
            "auto": True,
            "nullable": False,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int4",
        },
        "field_bigint": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int8",
        },
        "field_bigserial": {
            "pk": False,
            "auto": True,
            "nullable": False,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int8",
        },
        "field_bit": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "bit",
        },
        "field_varbit": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "varbit",
        },
        "field_box": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "box",
        },
        "field_bytea": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "bytea",
        },
        "field_char": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "bpchar",
        },
        "field_varchar": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "varchar",
        },
        "field_cidr": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "cidr",
        },
        "field_circle": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "circle",
        },
        "field_date": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "date",
        },
        "field_float8": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "float8",
        },
        "field_inet": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "inet",
        },
        "field_int4": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int4",
        },
        "field_interval": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "interval",
        },
        "field_json": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "json",
        },
        "field_jsonb": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "jsonb",
        },
        "field_line": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "line",
        },
        "field_lseg": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "lseg",
        },
        "field_macaddr": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "macaddr",
        },
        "field_macaddr8": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "macaddr8",
        },
        "field_money": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "money",
        },
        "field_numeric": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "numeric",
        },
        "field_path": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "path",
        },
        "field_point": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "point",
        },
        "field_polygon": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "polygon",
        },
        "field_float4": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "float4",
        },
        "field_int2": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int2",
        },
        "field_smallserial": {
            "pk": False,
            "auto": True,
            "nullable": False,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "smallserial",
        },
        "field_text": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "text",
        },
        "field_time": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "time",
        },
        "field_timetz": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "timetz",
        },
        "field_timestamp": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "timestamp",
        },
        "field_timestamptz": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "timestamptz",
        },
        "field_tsquery": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "tsquery",
        },
        "field_tsvector": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "tsvector",
        },
        "field_uuid": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "uuid",
        },
        "field_xml": {
            "pk": False,
            "auto": False,
            "nullable": True,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "xml",
        },
    }
    tablespec_rel_fieldnames = {
        "id_tablespec_rel": {
            "pk": True,
            "auto": True,
            "nullable": False,
            "fk": False,
            "fk_table": None,
            "fk_schema": None,
            "fk_column": None,
            "dtype": "int4",
        },
        "fk_tablespec_serial": {
            "pk": False,
            "auto": False,
            "nullable": False,
            "fk": True,
            "fk_table": "tablespec_serial",
            "fk_schema": "public",
            "fk_column": "id_tablespec",
            "dtype": "int4",
        },
    }

    def test_tablespec(self, pokie_db):
        generator = PgTableSpec(pokie_db)
        spec = generator.generate("tablespec_serial")
        assert spec.table == "tablespec_serial"
        assert spec.schema == "public"
        assert spec.pk == "id_tablespec"
        assert len(spec.fields) == len(self.tablespec_fieldnames)
        for field in spec.fields:
            assert field.name in self.tablespec_fieldnames.keys()
            field_spec = self.tablespec_fieldnames[field.name]
            assert field.pk == field_spec["pk"]
            assert field.auto == field_spec["auto"]
            assert field.nullable == field_spec["nullable"]
            assert field.fk == field_spec["fk"]
            assert field.fk_table == field_spec["fk_table"]
            assert field.fk_schema == field_spec["fk_schema"]
            assert field.fk_column == field_spec["fk_column"]

    def test_tablespec_rel(self, pokie_db):
        generator = PgTableSpec(pokie_db)
        spec = generator.generate("tablespec_rel")
        assert spec.table == "tablespec_rel"
        assert spec.schema == "public"
        assert spec.pk == "id_tablespec_rel"
        assert len(spec.fields) == len(self.tablespec_rel_fieldnames)
        for field in spec.fields:
            assert field.name in self.tablespec_rel_fieldnames.keys()
            field_spec = self.tablespec_rel_fieldnames[field.name]
            assert field.pk == field_spec["pk"]
            assert field.auto == field_spec["auto"]
            assert field.nullable == field_spec["nullable"]
            assert field.fk == field_spec["fk"]
            assert field.fk_table == field_spec["fk_table"]
            assert field.fk_schema == field_spec["fk_schema"]
            assert field.fk_column == field_spec["fk_column"]



from pokie.constants import HTTP_OK, HTTP_BADREQ, HTTP_NOT_FOUND
from pokie.rest import RestService
from pokie.test import PokieClient
from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


class TestRestView:
    base_url = "/customers"

    def test_view_get(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get(self.base_url)
            assert result.code == HTTP_OK
            assert result.success is True

            assert result.data["total"] > 50
            assert isinstance(result.data["items"], list)

            # get one item
            url = "{}/{}".format(self.base_url, "FIXTURE")
            result = client.get(url)
            assert result.code == HTTP_OK
            assert result.success is True

            assert isinstance(result.data, dict)
            for name in [
                "company_name",
                "contact_name",
                "contact_title",
                "address",
                "city",
                "region",
                "postal_code",
                "country",
                "phone",
                "fax",
            ]:
                assert name in result.data.keys()

    def test_view_post(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            record = CustomerRecord(
                id="TEST_",
                company_name="company_name",
                contact_name="contact_name",
                contact_title="contact_title",
                address="address",
                city="city",
                region="region",
                postal_code="pc",
                country="country",
                phone="phone",
                fax="fax",
            )

            # create new item
            result = client.post(self.base_url, data=record.asdict())
            assert result.code == HTTP_OK
            assert result.success is True

            # get one item
            url = "{}/{}".format(self.base_url, record.id)
            result = client.get(url)
            assert result.code == HTTP_OK
            assert result.success is True

            assert isinstance(result.data, dict)
            for name in [
                "company_name",
                "contact_name",
                "contact_title",
                "address",
                "city",
                "region",
                "postal_code",
                "country",
                "phone",
                "fax",
            ]:
                assert name in result.data.keys()

            # create invalid record
            record = CustomerRecord(
                id="FIXTURE",
                company_name="company_name",
                contact_name="contact_name",
                contact_title="contact_title",
                address="address",
                city="city",
                region="region",
                postal_code="pc",
                country="country",
                phone="phone",
                fax="fax",
            )

            # create new item with duplicate key
            result = client.post(self.base_url, data=record.asdict())
            assert result.code == HTTP_BADREQ
            assert result.success is False

    def test_view_put(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # get one item
            url = "{}/{}".format(self.base_url, "FIXTURE")
            result = client.get(url)
            assert result.code == HTTP_OK
            assert result.success is True

            record = result.data
            record["company_name"] = "UPDATED"
            # update item
            result = client.put(url, data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # get item again
            result = client.get(url)
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["company_name"] == "UPDATED"

            # update record with too long data
            record = result.data
            record["postal_code"] = "ABCDEFGHIJK"
            # update item
            result = client.put(url, data=record)
            assert result.code == HTTP_BADREQ
            assert result.success is False

    def test_view_delete(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            url = "{}/{}".format(self.base_url, "FIXTURE")
            result = client.delete(url)
            assert result.code == HTTP_OK
            assert result.success is True

            result = client.get(url)
            assert result.code == HTTP_NOT_FOUND
            assert result.success is False



import pickle

from rick.base import Di
from rick.mixin import Injectable
from rick.resource.redis import RedisCache as BaseRedisCache

from pokie.constants import DI_REDIS


class RedisCache(BaseRedisCache, Injectable):
    def __init__(self, di: Di):
        # base class init is completely overridden
        if not di.has(DI_REDIS):
            raise RuntimeError("DI_REDIS not found; maybe RedisFactory is missing?")
        self.set_di(di)
        self._serialize = pickle.dumps
        self._deserialize = pickle.loads
        self._prefix = None
        self._redis = di.get(DI_REDIS)



from io import StringIO


class TextBuffer:
    TAB_SIZE = 4

    def __init__(self, buffer: StringIO = None):
        self.buf = buffer
        if buffer is None:
            self.buf = StringIO()

    def tab(self, count=0):
        return " " * (count * self.TAB_SIZE)

    def newline(self, count=1):
        return "\n" * count

    def writeln(self, content, level=0, newlines=1):
        self.buf.write(self.tab(level))
        self.buf.write(content)
        self.buf.write(self.newline(newlines))

    def read(self) -> str:
        self.buf.seek(0)
        return self.buf.read()



from .user import UserCreateCmd, UserInfoCmd, UserModCmd, UserListCmd
from .acl import (
    AclRoleListCmd,
    AclRoleCreateCmd,
    AclRoleRemoveCmd,
    AclRoleInfoCmd,
    AclRoleLinkCmd,
    AclRoleUnlinkCmd,
)
from .acl import (
    AclResourceListCmd,
    AclResourceCreateCmd,
    AclResourceLinkCmd,
    AclResourceUnlinkCmd,
    AclUserRoleCmd,
)
